import os
import inspect
import pathlib
import requests
from metadrive import config
from metadrive import utils
from metadrive import mixins

SUBTOOL = os.path.basename(__file__).split('.py')[0]

def get_session(*args, **kwargs):
    return requests.Session(*args, **kwargs)

class RequestsDrive(requests.Session):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response = None
        self.desired_capabilities = {}
        self.metaname = ''

    def get(self, *args, **kwargs):

        proxy = self.desired_capabilities.get('proxy')
        if isinstance(proxy, dict):
            socks = proxy.get('socksProxy')
            if isinstance(socks, str):
                proxies = {
                    'http': 'socks5h://{}'.format(socks),
                    'https': 'socks5h://{}'.format(socks)}
                kwargs.update({'proxies': proxies})


        self.response = super().get(*args, **kwargs)

        if hasattr(self, 'profile'):
            session_data = requests.utils.dict_from_cookiejar(self.cookies)

            session_prefix_file = os.path.join(
                SUBTOOL, '{drive_id}/cookies.json'.format(
                    drive_id=self.profile
            ))

            utils.save_session_data(session_prefix_file, session_data)

    def quit(self):
        del self


def get_drive(
        profile='default',
        porfiles_dir='.metadrive/sessions/_requests',
        recreate_profile=False,
        proxies='default'):

    ## ----------- TO MOVE TO MIXIN --------------- #
    proxy = mixins.set_proxies(proxies)
    local = mixins.init_profile(profile, porfiles_dir, recreate_profile)

    # Instantiating Drive

    if local:
        drive = RequestsDrive()
        drive.subtool = SUBTOOL
        drive.profile = profile
    else:
        drive = None


    if drive is not None:

        session_prefix_file = os.path.join(
            drive.subtool, '{drive_id}/cookies.json'.format(
                drive_id=profile
        ))

        if os.path.exists(os.path.join(config.SESSIONS_DIR, session_prefix_file)):
            session_data = utils.load_session_data(session_prefix_file)
        else:
            session_data = requests.utils.dict_from_cookiejar(drive.cookies)
            # utils.save_session_data(session_prefix_file, session_data)

        if session_data:
            drive.cookies.update(
                requests.utils.cookiejar_from_dict(
                    session_data
                )
            )

        if proxy is not None:
            drive.desired_capabilities.update(
                {'proxy': proxy}
            )


        # LATER MOVE TO MIXIN
        drive.caller_module = inspect.getmodule(inspect.currentframe().f_back)

    return drive



from __future__ import with_statement

import os
import sys
import errno

from fuse import FUSE, FuseOSError, Operations


class Passthrough(Operations):
    def __init__(self, root):
        self.root = root

    # Helpers
    # =======

    def _full_path(self, partial):
        if partial.startswith("/"):
            partial = partial[1:]
        path = os.path.join(self.root, partial)
        return path

    # Filesystem methods
    # ==================

    def access(self, path, mode):
        full_path = self._full_path(path)
        if not os.access(full_path, mode):
            raise FuseOSError(errno.EACCES)

    def chmod(self, path, mode):
        full_path = self._full_path(path)
        return os.chmod(full_path, mode)

    def chown(self, path, uid, gid):
        full_path = self._full_path(path)
        return os.chown(full_path, uid, gid)

    def getattr(self, path, fh=None):
        full_path = self._full_path(path)
        st = os.lstat(full_path)
        return dict((key, getattr(st, key)) for key in ('st_atime', 'st_ctime',
                     'st_gid', 'st_mode', 'st_mtime', 'st_nlink', 'st_size', 'st_uid'))

    def readdir(self, path, fh):
        full_path = self._full_path(path)

        dirents = ['.', '..']
        if os.path.isdir(full_path):
            dirents.extend(os.listdir(full_path))
        for r in dirents:
            yield r

    def readlink(self, path):
        pathname = os.readlink(self._full_path(path))
        if pathname.startswith("/"):
            # Path name is absolute, sanitize it.
            return os.path.relpath(pathname, self.root)
        else:
            return pathname

    def mknod(self, path, mode, dev):
        return os.mknod(self._full_path(path), mode, dev)

    def rmdir(self, path):
        full_path = self._full_path(path)
        return os.rmdir(full_path)

    def mkdir(self, path, mode):
        return os.mkdir(self._full_path(path), mode)

    def statfs(self, path):
        full_path = self._full_path(path)
        stv = os.statvfs(full_path)
        return dict((key, getattr(stv, key)) for key in ('f_bavail', 'f_bfree',
            'f_blocks', 'f_bsize', 'f_favail', 'f_ffree', 'f_files', 'f_flag',
            'f_frsize', 'f_namemax'))

    def unlink(self, path):
        return os.unlink(self._full_path(path))

    def symlink(self, name, target):
        return os.symlink(target, self._full_path(name))

    def rename(self, old, new):
        return os.rename(self._full_path(old), self._full_path(new))

    def link(self, target, name):
        return os.link(self._full_path(name), self._full_path(target))

    def utimens(self, path, times=None):
        return os.utime(self._full_path(path), times)

    # File methods
    # ============

    def open(self, path, flags):
        full_path = self._full_path(path)
        return os.open(full_path, flags)

    def create(self, path, mode, fi=None):
        full_path = self._full_path(path)
        return os.open(full_path, os.O_WRONLY | os.O_CREAT, mode)

    def read(self, path, length, offset, fh):
        os.lseek(fh, offset, os.SEEK_SET)
        return os.read(fh, length)

    def write(self, path, buf, offset, fh):
        os.lseek(fh, offset, os.SEEK_SET)
        return os.write(fh, buf)

    def truncate(self, path, length, fh=None):
        full_path = self._full_path(path)
        with open(full_path, 'r+') as f:
            f.truncate(length)

    def flush(self, path, fh):
        return os.fsync(fh)

    def release(self, path, fh):
        return os.close(fh)

    def fsync(self, path, fdatasync, fh):
        return self.flush(path, fh)


def mount(root, mountpoint):
    FUSE(Passthrough(root), mountpoint, nothreads=True, foreground=True)

if __name__ == '__main__':
    mount(sys.argv[1], sys.argv[2])


# Copyright (c) 2016, Stavros Korokithakis
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# https://github.com/skorokithakis/python-fuse-sample
#



import os
# import xarray
import pandas
import inspect
from metadrive import mixins


SUBTOOL = os.path.basename(__file__).split('.py')[0]


class XarrayDrive:

    def __init__(self):
        self.desired_capabilities = {}
        self.metaname = ''
        self.df = None

    def read_csv(self, *args, **kwargs):
        self.df = pandas.read_csv(*args, **kwargs)


def get_drive(
        profile='default',
        porfiles_dir='.metadrive/sessions/_xarray',
        recreate_profile=False,
        proxies='default'):

    proxy = mixins.set_proxies(proxies)
    local = mixins.init_profile(profile, porfiles_dir, recreate_profile)

    drive = XarrayDrive()
    drive.subtool = SUBTOOL
    drive.profile = profile

    # LATER MOVE TO MIXIN
    drive.caller_module = inspect.getmodule(inspect.currentframe().f_back)

    return drive



import os
import pathlib
from metadrive import config

def set_proxies(proxies):

    if proxies == 'default':
        proxies = config.ENSURE_PROXIES()

    if not proxies:
        proxies = {
            'httpProxy': None,
            'sslProxy': None,
            'socksProxy': None
        }

    proxy = {'proxyType': 'MANUAL'}
    for key in proxies:
        if proxies[key] is not None:
            if key == 'http_proxy':
                Key = 'httpProxy'
            elif key == 'ssl_proxy':
                Key = 'sslProxy'
            elif key == 'socks_proxy':
                Key = 'socksProxy'
            else:
                Key = key

            if Key in ['http', 'https']:
                Value = proxies[key].rsplit('//', 1)[-1]
                proxy['socksProxy'] = Value
            else:
                proxy[Key] = proxies[key]

    if len(proxy) <= 1:
        proxy = None

    return proxy

def init_profile(profile, porfiles_dir, recreate_profile):

    local = True

    if local:
        profile_path = os.path.join(
            str(pathlib.Path.home()),
            os.path.join(porfiles_dir, profile))
    else:
        profile_path = None

    if profile_path is not None:

        if not os.path.exists(profile_path):
            os.makedirs(profile_path)

        elif recreate_profile:
            import shutil
            shutil.rmtree(profile_path)

    return local



import io
import os
import bs4
import tqdm
import yaml
import json
import time
import urllib
import tarfile
import requests

from metadrive.config import KNOWN_DRIVERS


def auto_discover(refresh=True):
    '''
    Goes over all PyPI packages in existence, which end with "-driver" in their name
    and returns those which have __site_url__ in their package __init__.py file, as
    a mapping enabling the discovery of drivers for Internet sites.
    '''

    if os.path.exists(KNOWN_DRIVERS) and not refresh:
        site_drivers = yaml.load(open(KNOWN_DRIVERS).read(), Loader=yaml.Loader)
        return site_drivers

    print("Downloading PyPI ...")
    response = requests.get('https://pypi.org/simple/', stream=True)

    f = io.BytesIO()
    total_length = response.headers.get('content-length')
    if total_length is None:
        f.write(response.content)
    else:
        total_length = 10000000 #int(total_length)

        with tqdm.tqdm(total=total_length) as pbar:
            for data in response.iter_content(chunk_size=4096):
                pbar.update(len(data))
                f.write(data)

    bindata = f.getvalue()
    print("Done.")
    print("Reading package URLs ...")
    soup = bs4.BeautifulSoup(bindata, 'html.parser')
    links = soup.find_all('a')

    # packages known to be not a metadrive drivers
    ignore = [
        'https://pypi.org/simple/aquasystems-driver/',
        'https://pypi.org/simple/arun-cassandra-driver/',
        'https://pypi.org/simple/dlogg-driver/',
        'https://pypi.org/simple/dse-driver/',
        'https://pypi.org/simple/moxel-http-driver/',
        'https://pypi.org/simple/moxel-python-driver/',
        'https://pypi.org/simple/oceandb-elasticsearch-driver/',
        'https://pypi.org/simple/oceandb-mongodb-driver/',
        'https://pypi.org/simple/openstack-vim-driver/',
        'https://pypi.org/simple/osmosis-aws-driver/',
        'https://pypi.org/simple/osmosis-azure-driver/',
        'https://pypi.org/simple/osmosis-on-premise-driver/',
        'https://pypi.org/simple/pg-driver/',
        'https://pypi.org/simple/sensirion-shdlc-driver/',
        'https://pypi.org/simple/slipstream-libcloud-driver/',
        'https://pypi.org/simple/supy-driver/',
        'https://pypi.org/simple/wolphin-driver/',
        'https://pypi.org/simple/yb-cassandra-driver/',
    ]

    site_drivers = []

    # This is for retrieving setup.py details.
    from metadrive.utils import stdoutIO
    import setuptools
    def setup(**kwargs):
        print(json.dumps(kwargs))
    setuptools.setup = setup

    print("Looking for driver packages and reading __init__.py files ...")
    for link in tqdm.tqdm(links):
        if link.attrs['href'].endswith('-driver/'):

            name = link.text
            link = urllib.parse.urljoin('https://pypi.org', link.attrs['href'])

            if link in ignore:
                continue
            response = requests.get(link)
            if response.ok:

                if response.text:
                    soup = bs4.BeautifulSoup(response.content, 'html.parser')
                    versions = soup.find_all('a')

                    if versions:
                        last_version = versions[-1]
                        response = requests.get(last_version.attrs['href'])

                        try:
                            tar = tarfile.open(mode= "r:gz", fileobj = io.BytesIO(response.content))
                        except:
                            tar = None

                        if tar is not None:
                            # import pdb; pdb.set_trace()
                            for member in tar.getnames():

                                if member.endswith('driver/__init__.py'):
                                    content = tar.extractfile(member).read()
                                    if content:
                                        text = content.decode('utf-8')
                                        if text:
                                            for line in text.split('\n'):
                                                if '__site_url__' in line:
                                                    if '=' in line:
                                                        site = line.split('=',1)[-1].strip()[1:-1]
                                                        if site:
                                                            if site.startswith('http'):
                                                                domain = urllib.parse.urlparse(site).hostname
                                                            else:
                                                                domain = None

                                                            for member in tar.getnames():
                                                                if member.count('/') == 1:
                                                                    if member.endswith('/setup.py'):
                                                                        setup = tar.extractfile(member).read()
                                                                        if setup:
                                                                            info = setup.decode('utf-8')
                                                                            with stdoutIO() as s:
                                                                                exec(info)
                                                                            info = json.loads(s.getvalue())
                                                                        else:
                                                                            info = None

                                                            site_drivers.append(
                                                                {'site_url': site,
                                                                 'domain': domain,
                                                                 'package': name,
                                                                 'type': 'pypi',
                                                                 'info': info})
                                                    break
                                    break

    with open(KNOWN_DRIVERS, 'w') as f:
        f.write(yaml.dump(site_drivers, Dumper=yaml.Dumper))

    print("Saved to {}. Now you can use metadrive.drivers.all() to quickly access them.".format(KNOWN_DRIVERS))

    return site_drivers

def index():
    return auto_discover(refresh=False)



from metadrive.tests.conftest import client

def test_sanity():
    assert 1 == 1

def test_root(client):
    response = client.get('http://0.0.0.0:8000')
    assert response.text.startswith('<!doctype ht')

def test_drivers(client):
    response = client.get('http://0.0.0.0:8000/drivers')
    assert type(response.json()) == list

def test_drives(client):
    response = client.get('http://0.0.0.0:8000/drives')
    assert type(response.json()) == list



import pytest
from starlette.config import environ
from starlette.testclient import TestClient

environ['TESTING'] = 'TRUE'

from metadrive.api import app

@pytest.fixture()
def client():
    with TestClient(app) as test_client:
        yield test_client



import metadrive

def test_read_table():
    drive = metadrive.drives.get('table-driver:default')
    from table_driver.api import Row

    expect = {'x': 1, 'y': 2, 'z': 3,
              '-': 'metadrive/tests/sample.csv#0',
              '@': 'PyPI::table-driver==0.0.3:default.api.Row'}

    result = Row._get('metadrive/tests/sample.csv#0', drive=drive)

    assert result['@'].startswith('PyPI::table-driver')
    assert result['@'].endswith('default.api.Row')

    del expect['@']
    del result['@']
    assert expect == result

